==> classforge-0.1.tar.gz: classforge-0.1/classforge/fields.py <==
# (C) Michael DeHaan <michaeldehaan.net>, 2020



class Field(object):

    __slots__ = [
        '_id',
        'name',
        'type',
        'default',
        'nullable',
        'mutable',
        'accessor',
        'mutator',
        '_mutated',
        'choices',
        'value'
    ]

    def __init__(self, name=None, type=None, default=None, nullable=True, mutable=True, mutator=None, accessor=None, choices=None):

        assert mutator is None or (isinstance(mutator, str))
        assert accessor is None or (isinstance(accessor, str))

        self._id = 0
        self.type = type
        self.default = default
        self.nullable = nullable
        self.mutable = mutable
        self.mutator = mutator
        self._mutated = False
        self.accessor = accessor
        self.choices = choices
        self.value = None

        if default is not None and self.value is None:
            self.value = default

    def show(self):
        return dict(
            type = self.type,
            default = self.default,
            nullable = self.nullable,
            mutable = self.mutable,
            mutator = self.mutator,
            accessor = self.accessor,
            choices = self.choices,
            value = self.value,
            _id = self._id,
        )

    def copy(self):
        obj = Field(type=self.type, default=self.default, nullable=self.nullable, mutable=self.mutable, mutator=self.mutator, accessor=self.accessor, choices=self.choices)
        obj._mutated = False
        obj.value = self.value
        obj._id = self._id + 1
        return obj

    def _INTERNAL_get(self, parent, name):
        fn = self._find_function(parent, self.accessor, self._access, "get_%s" % name)
        val = fn(self.value)
        return val


    def _INTERNAL_set(self, parent, name, value):

        self._mutated = True

        if not self.mutable and self._mutated:
            raise ValueError("field is immutable: %s" % name)

        if value is None and not self.nullable:
            raise ValueError("field is not nullable: %s" % name)

        if (self.nullable and value is not None) and (self.type is not None):
            if not isinstance(value, self.type):
                raise ValueError("invalid field type. got %s, expected: %s" % (value, self.type))

        if self.choices is not None:
            if not value in self.choices:
                raise ValueError("rejected value: %s" % name)

        fn = self._find_function(parent, self.mutator, self._mutate, "set_%s" % name)
        val = fn(value)
        self.value = val

    def _find_function(self, parent, member, default_func, func_pattern):

        if member is None:
            fn2 = getattr(parent, func_pattern, None)
            if fn2 is not None:
                return fn2
        else:
            fn3 = getattr(parent, member, None)
            if fn3 is None:
                raise AttributeError("missing function: %s" % member)
            return fn3
        return default_func

    def _access(self, value):
        return value

    def _mutate(self, value):
        return value

==> classforge-0.2.tar.gz: classforge-0.2/classforge/fields.py <==
# (C) Michael DeHaan <michaeldehaan.net>, 2020



class Field(object):

    __slots__ = [
        'name',
        'type',
        'default',
        'nullable',
        'mutable',
        'accessor',
        'mutator',
        '_mutated',
        'choices',
        'value',
        'required',
        'remap',
        'hidden'
    ]

    def __init__(self, name=None, type=None, default=None, nullable=True, mutable=True, mutator=None, accessor=None, choices=None, required=False, hidden=False, remap=None):

        assert mutator is None or (isinstance(mutator, str))
        assert accessor is None or (isinstance(accessor, str))

        self.type = type
        self.default = default
        self.nullable = nullable
        self.mutable = mutable
        self.mutator = mutator
        self._mutated = False
        self.accessor = accessor
        self.choices = choices
        self.value = None
        self.required = required
        self.hidden = hidden
        self.remap = remap

        if default is not None and self.value is None:
            self.value = default

    def _show(self):

        # this is just used for debugging, not part of a public API
        return dict(
            type = self.type,
            default = self.default,
            nullable = self.nullable,
            mutable = self.mutable,
            mutator = self.mutator,
            accessor = self.accessor,
            choices = self.choices,
            value = self.value,
            required = self.required,
            remap = self.remap,
            hidden = self.hidden
        )

    def copy(self):
        obj = Field(type=self.type, default=self.default, nullable=self.nullable, mutable=self.mutable, mutator=self.mutator, accessor=self.accessor, choices=self.choices, required=self.required, remap=self.remap, hidden=self.hidden)
        obj._mutated = False
        obj.value = self.value
        return obj

    def _INTERNAL_get(self, parent, name):
        # return the value of the field
        fn = self._find_function(parent, self.accessor, self._access, "get_%s" % name)
        val = fn(self.value)
        return val

    def _INTERNAL_set(self, parent, name, value):
        # set the value of the field
        self._mutated = True

        if not self.mutable and self._mutated:
            raise ValueError("field is immutable: %s" % name)

        if value is None and not self.nullable:
            raise ValueError("field is not nullable: %s" % name)

        if (self.nullable and value is not None) and (self.type is not None):
            if not isinstance(value, self.type):
                raise ValueError("invalid field type for %s. got %s, expected: %s" % (name, value, self.type))

        if self.choices is not None:
            if not value in self.choices:
                raise ValueError("rejected value for %s: choices=%s" % (name, self.choices))

        fn = self._find_function(parent, self.mutator, self._mutate, "set_%s" % name)
        val = fn(value)
        self.value = val

    def _find_function(self, parent, member, default_func, func_pattern):

        if member is None:
            # no named mutator function was passed into the field, so...
            fn2 = getattr(parent, func_pattern, None)
            if fn2 is not None:
                # use a funciton named get/set_<fieldname> if available...
                return fn2
        else:
            # a named member function was specified, so use if available, or error
            fn3 = getattr(parent, member, None)
            if fn3 is None:
                raise AttributeError("missing function: %s" % member)
            return fn3
        return default_func

    def _access(self, value):
        # optionally could override this in a subclass
        return value

    def _mutate(self, value):
        # optionally could override this in a subclass
        return value
